/*
** Module   : ClientBuilder.java
** Abstract : build and start a platform process.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------Description------------------------------------
** 001 MAG 20131118 First version.
** 002 MAG 20140206 Send the referrer address via command line.
** 003 CA  20140206 Refactored to make it unaware of the type of the P2J client being spawned.
** 004 CA  20140224 Redirect STDERR of spawned process to the server STDERR.
** 005 MAG 20140310 Set password.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.util.*;

/**
 * This class is used to create and spawn a client process using a
 * {@link ProcessBuilder} instance. A special small tools are designed for
 * Linux / Windows OS in order to spawn the process on the OS user account.
 * Both tools are written in C using specific OS API and are build as executable
 * using ant native target. The source code and the make file for this tools are 
 * available in the /src/native process directory. 
 * The password is provided vis stdin redirection on both platforms. 
 * The command line parameters for the spawned process are constructed from default 
 * parameters stored in directory and some specific parameters stored inside a 
 * <code>ClientBuilderParameters</code> structure.   
 * 
 * @author mag
 *
 */
public abstract class ClientBuilder
{   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ClientBuilder.class.getName());

   /** Pending spawn commands to be retrieved by remote clients. */
   private static final Map<String, String[]> pendingCmds = new HashMap<>();
   
   /** A collection of command line tokens */
   protected List<String> command = new LinkedList<String>();
      
   /** A Java Bean containing specific parameters */
   protected final ClientBuilderParameters  config;
   
   /** The ProcessBuilder instance */
   private ProcessBuilder pb = new ProcessBuilder();
   
   /**
    * Constructor.
    * 
    * @param   config
    *          The specific parameters for this client.
    */
   public ClientBuilder(ClientBuilderParameters config)
   {
      this.config = config;
   }

   /**
    * Get the spawn command registered for this UUID.
    * 
    * @param    uuid
    *           The UUID.
    *           
    * @return   See above.
    */
   public static String[] getCommand(String uuid)
   {
      synchronized (pendingCmds)
      {
         return pendingCmds.get(uuid);
      }
   }
   
   /**
    * Remove the spawn command registered for this UUID.
    * 
    * @param    uuid
    *           The UUID.
    */
   public static void removeCommand(String uuid)
   {
      synchronized (pendingCmds)
      {
         pendingCmds.remove(uuid);
      }
   }
   
   /**
    * Add custom client options to the {@link #command}.
    */
   protected abstract void addClientOptions();
   
   /**
    * Get the spawn arguments for a specific implementation.
    * 
    * @return   See above.
    */
   protected abstract List<String> getSpawnArguments();

   /**
    * Spawn's new Java process.
    * 
    * @param   env
    *          A map with environment variables for temporary account.
    *           
    * @return  Process object.
    * 
    * @throws  IOException
    *          If something goes wrong.
    */
   public Process start(Map<String, String> env) throws IOException
   {
      SessionManager sessMgr = SessionManager.get();
      BootstrapConfig cfg    = sessMgr.config();
      int securePort         = cfg.getInt("net", "server", "secure_port", -1);
      if (securePort == -1)
      {
         throw new RuntimeException(
                          "Clients can be spawned only if the server is running in secure mode!");
      }
      
      // build the final command (which will be directly started or delayed until after P2J
      // authentication
      {
         String userDir = config.getWorkingDirectory();
   
         // User name.
         command.add(config.getUserName());
         // Password.
         if (!config.isPasswordAuthentication())
         {
            String password = config.getPassword();
            command.add(password == null ? "\"\"" : password);
         }
         // Working directory
         command.add(userDir);
         // Build command
         command.add(config.getCommand());
         // Add JVM arguments
         command.addAll(config.getJvmArguments());
         // Add CLASSPATH and main class
         addClasspath();
         // Add client configuration file
         addConfigFile(userDir);
         // Add server options
         addServerOptions(config.isSecure());
         // Add other client options
         addClientOptions();
         // Add custom bootstrap configuration
         command.addAll(config.getBootstrapConfig());
         // Add application arguments
         List<String> appArgs = config.getAppArguments();
         if (appArgs != null && appArgs.size() > 0)
         {
            boolean first = true;
            Iterator<String> iter = appArgs.iterator();
            while (iter.hasNext())
            {
               String appArg = iter.next().trim();
               if (appArg.length() > 0)
               {
                  if (first)
                  {
                     command.add("-a");
                     first = false;
                  }
                  command.add(appArg);
               }
            }
         }
   
         // ensure there are no empty or null arguments. an empty argument will make the 
         // ClientDriver to assume that is a bootstrap config file name and fail on startup.
         Iterator<String> iter = command.iterator();
         while (iter.hasNext())
         {
            String s = iter.next();
            if (s == null || s.trim().length() == 0)
            {
               iter.remove();
            }
         }
         
         if (!config.isPasswordAuthentication())
         {
            synchronized (pendingCmds)
            {
               // save this for later use
               pendingCmds.put(config.getUuid(), command.toArray(new String[0]));
            }
         }
      }
       
      // launch the spawner
      File spawnerDir = new File(config.getSpawner());
      List<String> spawnCmd = new ArrayList<>();
      spawnCmd.add(config.getSpawner());
      spawnCmd.addAll(getSpawnArguments());

      if (LOG.isLoggable(Level.FINE))
      {
         StringBuilder sb = new StringBuilder();
         for (String s : spawnCmd)
         {
            sb.append(s).append(" ");
         }
         
         final String msg = "Launching a new P2J client using the command: %s";
         LOG.logp(Level.FINE, "ClientBuilder", "start", String.format(msg, sb.toString().trim()));
      }

      pb.environment().putAll(env);
      // we start in the spawner's dir
      pb.directory(spawnerDir.getAbsoluteFile().getParentFile());
      pb.command(spawnCmd);
      pb.redirectError(ProcessBuilder.Redirect.INHERIT);
      
      Process shell = pb.start();

      if (config.isPasswordAuthentication())
      {
         // write password
         OutputStreamWriter os = new OutputStreamWriter(shell.getOutputStream());
         os.write(config.getPassword());
         os.close();
      }
      
      return shell;
   }

   /**
    * Add server properties.
    * 
    * @param   secure
    *          Indicating to setup a secure <code>true</true> 
    *          or an insecure <code>false</code> connection.
    */
   private void addServerOptions(boolean secure)
   {
      // uuid assigned to this client
      command.add("server:spawner:uuid=" + config.getUuid());
      
      // add target server ports
      BootstrapConfig serverConfig = SessionManager.get().config();

      String host = serverConfig.getString("net", "server", "host", "localhost");
      command.add("net:server:host=" + host);

      if (secure)
      {
         command.add("net:connection:secure=true");
         getServerConfig(serverConfig, "net", "server", "secure_port", -1);
      }
      else if (getServerConfig(serverConfig, "net", "server", "port", -1) == -1)
      {
         getServerConfig(serverConfig, "net", "server", "insecure_port", -1);
      }
   }

   /**
    * Get a key value form configuration file. 
    * If key exists add the value to builder command.
    * 
    * @param   config
    *          Configuration file.
    * @param   cname
    *          Category name.
    * @param   gname
    *          Group name.
    * @param   kname
    *          Key name.
    * @param   def
    *          Key value if exists default value otherwise.
    * 
    * @return  The integer value for key.
    */
   private int getServerConfig(BootstrapConfig config,
                               String cname, 
                               String gname, 
                               String kname,
                               int def)
   {
      int value = config.getInt(cname, gname, kname, def);
      if (value != def)
      {
         command.add(String.format("%s:%s:%s=%d", cname, gname, kname, value));
      }
      return value;
   }

   /**
    * Add Java CLASSPATH and the main class.
    * 
    */
   private void addClasspath()
   {
      Class<?> mainClass = config.getMainClass();
      // Native libraries path
      String libPath = config.getLibPath();
      command.add(String.format("-Djava.library.path=%s", libPath));
      // Add CLASSPATH
      command.add("-classpath");
      command.add(config.getClassPath());
      // Add main class
      command.add(mainClass.getName());
   }

   /**
    * Add client configuration file if exists.
    * 
    * @param   userDir
    *          Client working directory.
    */
   private void addConfigFile(String userDir)
   {
      String configFile = config.getConfigFile();
      if (configFile != null)
      {
         String clientConfig = String.format("%s%s%s", userDir, File.separator, configFile);
         if (new File(clientConfig).exists())
         {
            command.add(configFile);
         } 
         else
         {
            LOG.logp(Level.WARNING, 
                     "ClientBuilder.addConfigFile()", 
                     "", 
                     LogHelper.generate("%s does not exist.", clientConfig));
         }
      }
   }

   /**
    * Sets this process builder's working directory.
    * 
    * @param   userDir
    *          Server working directory.
    */
   @SuppressWarnings("unused")
   private File setUserDir(String userDir)
   {
      File fUserDir = new File(userDir);
      if (!fUserDir.exists())
      {
         String msg = "The user dir '%s' specified does not exist.";
         throw new IllegalStateException(String.format(msg, userDir));
      }
      return fUserDir;
   }
}
